@@ -1,10 +1,8 @@
-from datetime import datetime
 from test.conftest import USE_MOCK
 
 import pytest
 
 import vessl
-from vessl.util.constant import VESSL_PLOTS_FILETYPE_IMAGE
 from vessl.util.exception import InvalidExperimentError
 
 
