@@ -1,7 +1,7 @@
 from typing import List
 
-import keras
 import numpy as np
+from tensorflow import keras
 
 from vessl.experiment import list_experiments
 from vessl.integration.common import get_current_experiment
