@@ -5,7 +5,7 @@ import signal
 import sys
 import threading
 import time
-from typing import Any, Dict, List, Optional
+from typing import Any, Dict, List, Optional, Union
 
 from urllib3.exceptions import MaxRetryError
 
@@ -13,24 +13,32 @@ from openapi_client.models.experiment_metric_entry import ExperimentMetricEntry
 from openapi_client.models.experiment_metrics_update_api_payload import (
     ExperimentMetricsUpdateAPIPayload,
 )
-from openapi_client.models.experiment_progress_update_api_payload import (
-    ExperimentProgressUpdateAPIPayload,
-)
 from openapi_client.models.local_experiment_finish_api_payload import (
     LocalExperimentFinishAPIPayload,
 )
 from openapi_client.models.response_experiment_info import ResponseExperimentInfo
 from vessl.cli._util import Endpoint
+from vessl.integration.tensorboard import integrate_tensorboard
 from vessl.internal.collector import (
     Collector,
     IOCollector,
     SystemMetricCollector,
     UserMetricCollector,
 )
+from vessl.internal.progress_updater import ProgressUpdater
 from vessl.util import logger
 from vessl.util.api import VesslApi
-from vessl.util.constant import VESSL_IMAGE_PATH, VESSL_PLOTS_FILETYPE_IMAGE
-from vessl.util.exception import InvalidExperimentError, VesslApiException
+from vessl.util.audio import Audio
+from vessl.util.constant import (
+    VESSL_MEDIA_PATH,
+    VESSL_PLOTS_FILETYPE_AUDIO,
+    VESSL_PLOTS_FILETYPE_IMAGE,
+)
+from vessl.util.exception import (
+    InvalidExperimentError,
+    InvalidTypeError,
+    VesslApiException,
+)
 from vessl.util.image import Image
 
 MODE_TEST = "TEST"
@@ -39,7 +47,6 @@ MODE_LOCAL = "LOCAL"
 MODE_MANAGED = "MANAGED"
 
 SEND_INTERVAL_IN_SEC = 10
-PROGRESS_UPDATE_INTERVAL_IN_SEC = 1
 
 METRIC_KEY_REGEX = re.compile("^[a-zA-Z0-9/_-]+$")
 
@@ -96,52 +103,6 @@ class Sender(object):
             logger.exception("Unexpected error", exc_info=e)
 
 
-class ProgressUpdater(object):
-    def __init__(self, api: VesslApi, experiment_id: int):
-        self._api = api
-        self._experiment_id = experiment_id
-        self._thread = threading.Thread(target=self._thread_body, daemon=True)
-        self._exit = threading.Event()
-
-        self._progress = None
-
-    def start(self):
-        self._thread.start()
-
-    def stop(self):
-        self._exit.set()
-        self._thread.join()
-
-    def update(self, value: float):
-        self._progress = value
-
-    def _thread_body(self):
-        while not self._exit.is_set():
-            self._send()
-            self._exit.wait(timeout=PROGRESS_UPDATE_INTERVAL_IN_SEC)
-        self._send()
-
-    def _send(self):
-        if self._progress is None:
-            return
-
-        logger.debug(f"Sending experiment progress: {self._progress}")
-        try:
-            self._api.experiment_progress_update_api(
-                self._experiment_id,
-                experiment_progress_update_api_payload=ExperimentProgressUpdateAPIPayload(
-                    progress_percent=self._progress,
-                ),
-            )
-            self._progress = None  # Flush after sending
-
-        except (MaxRetryError, VesslApiException) as e:
-            logger.exception("Failed to send metrics to server", exc_info=e)
-
-        except Exception as e:
-            logger.exception("Unexpected error", exc_info=e)
-
-
 class VesslRun(object):
     class ExitHook(object):
         def __init__(self, orig_exit):
@@ -162,6 +123,7 @@ class VesslRun(object):
         "_logger",
         "_user_metric_collector",
         "_exit_hook",
+        "_tensorboard_collector",
     ]
 
     def __init__(self) -> None:
@@ -261,7 +223,13 @@ class VesslRun(object):
         if res.rejected:
             logger.warning(f"{res.rejected} payloads(s) were rejected.")
 
-    def init(self, experiment_name_or_number=None, message: str = None, is_test=False):
+    def init(
+        self,
+        experiment_name_or_number=None,
+        message: str = None,
+        tensorboard: bool = False,
+        **kwargs,
+    ):
         """Main function to setup Vessl in a local setting
 
         If this is a Vessl-managed experiment or vessl.init has already been called,
@@ -270,35 +238,39 @@ class VesslRun(object):
         Args:
             experiment_name_or_number (str | int): experiment name or number
             message (str): experiment message
-            is_test (bool): internal use only
+            tensorboard (bool): enable tensorboard integration. It is important to note
+              that `vessl.init` must be called **before** initializing the writer
+              (`tf.summary.create_file_writer` for TF2, SummaryWriter for PyTorch, etc).
         """
-        if self._mode != MODE_NOT_STARTED:
-            return
+        if self._mode == MODE_NOT_STARTED:
+            print(f"Initializing a new experiment...")
+            self.api.configure()
+            self._experiment = self._get_experiment_from_args(
+                experiment_name_or_number, message
+            )
+            self._mode = MODE_LOCAL
 
-        print(f"Initializing a new experiment...")
-        self.api.configure()
-        self._experiment = self._get_experiment_from_args(
-            experiment_name_or_number, message
-        )
-        self._mode = MODE_LOCAL
+            print(
+                f"Connected to {self._experiment.name}.\n"
+                f"For more info: {Endpoint.experiment.format(self._experiment.organization.name, self._experiment.project.name, self._experiment.number)}"
+            )
 
-        print(
-            f"Connected to {self._experiment.name}.\n"
-            f"For more info: {Endpoint.experiment.format(self._experiment.organization.name, self._experiment.project.name, self._experiment.number)}"
-        )
+            # For testing
+            if "is_test" in kwargs and kwargs["is_test"]:
+                return
 
-        if is_test:
-            return
+            gpu_count = self._experiment.local_execution_spec.gpu_count or 0
+            self._user_metric_collector = UserMetricCollector()
+            collectors = [
+                IOCollector(),
+                SystemMetricCollector(gpu_count),
+                self._user_metric_collector,
+            ]
+            self._sender = Sender(self.api, self._experiment.id, collectors)
+            self._start()
 
-        gpu_count = self._experiment.local_execution_spec.gpu_count or 0
-        self._user_metric_collector = UserMetricCollector()
-        collectors = [
-            IOCollector(),
-            SystemMetricCollector(gpu_count),
-            self._user_metric_collector,
-        ]
-        self._sender = Sender(self.api, self._experiment.id, collectors)
-        self._start()
+        if tensorboard:
+            self._tensorboard_collector = integrate_tensorboard()
 
     def upload(self, path: str):
         """Upload output files
@@ -336,7 +308,12 @@ class VesslRun(object):
         self._experiment = None
         print(f"Experiment {experiment_name} completed.")
 
-    def log(self, payload: Dict[str, Any], step: Optional[int] = None):
+    def log(
+        self,
+        payload: Dict[str, Any],
+        step: Optional[int] = None,
+        ts: Optional[float] = None,
+    ):
         """Log metrics to Vessl
 
         Args:
@@ -348,14 +325,20 @@ class VesslRun(object):
             logger.warning("Invalid. Use `vessl.init()` first.")
             return
 
+        if ts is None:
+            ts = time.time()
+
         scalar_dict = {}
-        image_dict = {}
+        media_dict = {}
 
         for k, v in payload.items():
-            if isinstance(v, list) and all(isinstance(i, Image) for i in v):
-                image_dict[k] = v
-            elif isinstance(v, Image):
-                image_dict[k] = [v]
+            if isinstance(v, list):
+                if all(isinstance(i, Image) for i in v) or all(
+                    isinstance(i, Audio) for i in v
+                ):
+                    media_dict[k] = v
+            elif isinstance(v, Image) or isinstance(v, Audio):
+                media_dict[k] = [v]
             else:
                 scalar_dict[k] = v
 
@@ -364,11 +347,11 @@ class VesslRun(object):
         if scalar_dict or step is not None:
             self._user_metric_collector.handle_step(step)
 
-        if image_dict:
-            self._update_images(image_dict)
+        if media_dict:
+            self._update_media(media_dict, ts)
 
         if scalar_dict:
-            self._update_metrics(scalar_dict)
+            self._update_metrics(scalar_dict, ts)
 
     # This should mirror app/influx/metric_schema.go > `isValidMetricKey`
     def _is_metric_key_valid(self, key: str):
@@ -381,67 +364,80 @@ class VesslRun(object):
                 return False
         return True
 
-    def _update_metrics(self, payload: Dict[str, Any]):
+    def _update_metrics(self, payload: Dict[str, Any], ts: float):
         invalid_keys = [k for k in payload.keys() if not self._is_metric_key_valid(k)]
         if invalid_keys:
             logger.warning(
                 f"Invalid metric keys: {' '.join(invalid_keys)}. This payload will be rejected."
             )
 
-        payloads = [self._user_metric_collector.build_metric_payload(payload)]
+        payloads = [self._user_metric_collector.build_metric_payload(payload, ts)]
         if self._mode == MODE_MANAGED:
             self._send_without_collector(payloads)
             return self._user_metric_collector.step
         return self._user_metric_collector.log_metrics(payloads)
 
-    def _update_images(self, payload: Dict[str, List[Image]]):
+    def _update_media(
+        self,
+        payload: Dict[str, Union[List[Image], List[Audio]]],
+        ts: float,
+    ):
+        media_type, set_media_type = None, False
         path_to_caption = {}
-        for images in payload.values():
-            for image in images:
-                path_to_caption[image.path] = image.caption
-
-        source_path = os.path.join(VESSL_IMAGE_PATH, "")
-        assert self._experiment
-        dest_volume_id = self._experiment.experiment_plot_volume
-        dest_path = "/"
+        for media in payload.values():
+            for medium in media:
+                path_to_caption[medium.path] = medium.caption
+                if not set_media_type:
+                    if not isinstance(medium, Image) and not isinstance(medium, Audio):
+                        raise InvalidTypeError(
+                            f"Invalid payload type error: {type(medium)}"
+                        )
+                    else:
+                        media_type = type(medium).__name__
+                        set_media_type = True
 
         from vessl.volume import copy_volume_file
 
         files = copy_volume_file(
             source_volume_id=None,
-            source_path=source_path,
-            dest_volume_id=dest_volume_id,
-            dest_path=dest_path,
+            source_path=os.path.join(VESSL_MEDIA_PATH, ""),
+            dest_volume_id=self._experiment.experiment_plot_volume,
+            dest_path="/",
             recursive=True,
         )
 
-        for images in payload.values():
-            for image in images:
-                image.flush()
+        for media in payload.values():
+            for medium in media:
+                medium.flush()
 
+        plot_files = []
         if files:
             plot_files = [
                 {
                     "step": None,
                     "path": file.path,
                     "caption": path_to_caption[file.path],
-                    "timestamp": time.time(),
+                    "timestamp": ts,
                 }
                 for file in files
                 if file.path in path_to_caption
             ]
-        else:
-            plot_files = []
 
         payloads: List[ExperimentMetricEntry] = []
         for f in plot_files:
-            payload = {VESSL_PLOTS_FILETYPE_IMAGE: f}
-            payloads.append(self._user_metric_collector.build_image_payload(payload))
+            payload = {}
+            if media_type == Image.__name__:
+                payload = {VESSL_PLOTS_FILETYPE_IMAGE: f}
+            elif media_type == Audio.__name__:
+                payload = {VESSL_PLOTS_FILETYPE_AUDIO: f}
+            payloads.append(
+                self._user_metric_collector.build_media_payload(payload, ts)
+            )
 
         if self._mode == MODE_MANAGED:
             self._send_without_collector(payloads)
         else:
-            self._user_metric_collector.log_images(payloads)
+            self._user_metric_collector.log_media(payloads)
 
     def progress(self, value: float):
         """Update experiment progress
